/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.modules.declarative.rdf;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.jactr.core.chunk.IChunk;
import org.jactr.core.chunktype.IChunkType;
import org.jactr.core.chunktype.ISymbolicChunkType;
import org.jactr.core.module.declarative.six.DefaultDeclarativeModule6;
import org.jactr.core.slot.BasicSlot;

import org.openrdf.model.Namespace;
import org.openrdf.model.Resource;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.impl.URIImpl;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryConnection;
import org.openrdf.repository.RepositoryException;
import org.openrdf.repository.RepositoryResult;
import org.openrdf.repository.sail.SailRepository;
import org.openrdf.rio.RDFFormat;
import org.openrdf.sail.inferencer.fc.ForwardChainingRDFSInferencer;
import org.openrdf.sail.memory.MemoryStore;

import eu.larkc.core.data.CloseableIterator;
import eu.larkc.core.data.DataFactory;
import eu.larkc.core.data.LabelledGroupOfStatements;
import eu.larkc.core.data.RdfStoreConnection;
import eu.larkc.core.query.SPARQLQuery;
import eu.larkc.jactr.core.chunk.rdf.DefaultChunkRdf;
import eu.larkc.jactr.core.chunk.rdf.ISymbolicChunkRdf;
import eu.larkc.jactr.modules.declarative.rdf.DefaultDeclarativeModuleRdf;
import eu.larkc.jactr.modules.declarative.rdf.IDeclarativeModuleRdf;

/**
 * Default Declarative Module using an OpenRDF Sesame Repository.
 * 
 * @author Martijn Brakenhoff
 * @created March 15, 2010
 */
public class DefaultDeclarativeModuleRdf extends DefaultDeclarativeModule6 implements IDeclarativeModuleRdf {

	private static final Log	LOGGER	= LogFactory.getLog(DefaultDeclarativeModuleRdf.class);
	private static final URI	RDF_TYPE_URI 	= new URIImpl("http://www.w3.org/1999/02/22-rdf-syntax-ns#type");
	private static final URI	RDFS_DOMAIN_URI	= new URIImpl("http://www.w3.org/2000/01/rdf-schema#domain");
	private static final Value	RDFS_CLASS_URI	= (Value) new URIImpl("http://www.w3.org/2000/01/rdf-schema#Class");
	private static final String RDF_PREFIX 	= "rdf";
	private static final String RDFS_PREFIX = "rdfs";
	private static final String OWL_PREFIX 	= "owl";
	private static final String XSD_PREFIX 	= "xsd";
	private static final String BASE_URI_ERROR = "Unable to initialize the DeclarativeModuleRdf without a BaseURI parameter.";
	
	// Variables
	private static int 				_uniqueLabelID = 0;
	protected String				_baseURI;
	protected String				_sourceXML;
	protected String				_sourceFile; 
	protected String				_sourceURL;
	protected RDFFormat				_rdfFormat;
	protected Map<String, String> 	_namespaces;
	protected String 				_sparqlNamespaces;

	public DefaultDeclarativeModuleRdf() {
		_rdfFormat = RDFFormat.RDFXML;
		_namespaces = new HashMap<String, String>();
	}

	// +-------------------+
	// | Module Management |
	// +-------------------+

	@Override
	synchronized public void dispose() {
		super.dispose();
	}

	@Override
	public void initialize() {
		if (_baseURI == null) {
			if (LOGGER.isErrorEnabled()) {
				LOGGER.error(BASE_URI_ERROR);
			}
			throw new IllegalStateException(BASE_URI_ERROR);
		}
		
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Declarative Module Initialized");
		}
		super.initialize();

		ArrayList<Statement> rdf = new ArrayList<Statement>();

		if (_sourceURL != null) {
			try {			
				SailRepository repository = new SailRepository(
						new ForwardChainingRDFSInferencer(
								new MemoryStore()));

				repository.initialize();

				RepositoryConnection con = repository.getConnection();

				File file = new File(_sourceURL);
				File modelDirFile = new File("../../../models/" + _sourceURL);
				if (file.exists()) {	
					addLocalFileToRepository(repository, file);
				} else if (modelDirFile.exists()) {
					addLocalFileToRepository(repository, modelDirFile);
				} else {
					addURLToRepository(repository, new URL(_sourceURL));
				}

				con.getStatements(null, null, null, false).addTo(rdf);				

				setNamespaces( con.getNamespaces() );

				con.close();
			} catch (MalformedURLException e) {
				LOGGER.fatal("File/URL '"+ _sourceURL +"' was not found.");
			} catch (RepositoryException e) {
				LOGGER.fatal("Error while initializing Declarative Module: ", e);
			}
		}

		RdfStoreConnection localStore = DataFactory.INSTANCE.createRdfStoreConnection();

		// add the rdf to the data-layer.
		LabelledGroupOfStatements label = localStore.createLabelledGroupOfStatements();		
		for (Statement s : rdf) {
			localStore.addStatement(s.getSubject(), s.getPredicate(), s
					.getObject(), label.getLabel(), label.getLabel());
		}

		initiateChunkTypes();
	}

	private void addLocalFileToRepository ( Repository repository, File file ) {
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Adding '" + file.getName() + "' to the repository:");
		}

		try {
			RepositoryConnection con = repository.getConnection();

			con.add(file, _baseURI, _rdfFormat);

			con.close();
		} catch (Exception e) {
			LOGGER.error("Error while adding data to the repository: ", e);
		}
	}

	private void addURLToRepository ( Repository repository, URL url ) {
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Adding '" + url + "' to the repository:");
		}

		try {
			RepositoryConnection con = repository.getConnection();

			con.add(url, _baseURI, _rdfFormat);

			con.close();
		} catch (Exception e) {
			LOGGER.error("Error while adding data to the repository: ", e);
		}
	}

	protected void initiateChunkTypes () {
		RdfStoreConnection con = DataFactory.INSTANCE.createRdfStoreConnection();
		CloseableIterator<Statement> chunkTypeStatements = con.search(null, RDF_TYPE_URI, RDFS_CLASS_URI, null, null);

		while (chunkTypeStatements.hasNext()) {
			Statement chunkTypeStatement = chunkTypeStatements.next();

			Resource chunkTypeResource = chunkTypeStatement.getSubject();
			String name = addPrefix(chunkTypeResource.stringValue());

			if ( !name.startsWith( RDF_PREFIX ) &&
					!name.startsWith( RDFS_PREFIX ) &&
					!name.startsWith( OWL_PREFIX ) && 
					!name.startsWith( XSD_PREFIX ) ) {
				addChunkType( rdfToChunkType ( con, chunkTypeResource ) );

				if (LOGGER.isDebugEnabled()) { 
					LOGGER.debug("Added ChunkType: " + name);
				}
			}
		}
	}

	protected IChunkType rdfToChunkType (RdfStoreConnection con, Resource chunkTypeResource) {
		String chunkTypeName = addPrefix(chunkTypeResource.stringValue());
		IChunkType result = null;

		try {
			result = createChunkType( null, chunkTypeName ).get();
		} catch (Exception e) {
			LOGGER.debug( e.getMessage() );
		}

		ISymbolicChunkType sct = result.getSymbolicChunkType();

		CloseableIterator<Statement> chunkTypeStatements = con.search(null, RDFS_DOMAIN_URI, chunkTypeResource, null, null);
		while (chunkTypeStatements.hasNext()) {
			Statement statement = chunkTypeStatements.next();			
			Resource slotResource = statement.getSubject();

			sct.addSlot( new BasicSlot( addPrefix( slotResource.stringValue() ) ) );
		}

		return result;
	}

	// +------------+
	// | Namespaces |
	// +------------+

	private void setNamespaces(RepositoryResult<Namespace> namespaces) throws RepositoryException {
		_namespaces = new HashMap<String, String>();
		StringBuffer sb = new StringBuffer();

		while (namespaces.hasNext()) {
			Namespace ns = namespaces.next();
			_namespaces.put(ns.getPrefix(), ns.getName());

			String prefix = "PREFIX "+ ns.getPrefix() + ": <" + ns.getName() +">\n";
			sb.append(prefix);
		}
		_sparqlNamespaces = sb.toString();
	}

	public Map<String, String> getNamespaces () {
		return Collections.unmodifiableMap(_namespaces);
	}

	public String getSPARQLPrefixes () {
		return _sparqlNamespaces;
	}

	public String addPrefix(String resource) {
		Iterator<Entry<String, String>> entries = _namespaces.entrySet().iterator();

		while (entries.hasNext()) {
			Map.Entry<String, String> entry = entries.next();

			resource = resource.replace(entry.getValue(), entry.getKey()+":");
		}

		return resource;
	}

	// +----------------+
	// | Chunk Creation |
	// +----------------+

	/**
	 * This DeclarativeModule creates Rdf Chunks. Called by the superclass. 
	 * 
	 * @param parent The chunktype.
	 * @return
	 */
	@Override
	protected IChunk instantiateChunk(IChunkType parent) {
		return new DefaultChunkRdf(parent);
	}

	// +-----------------+
	// | SPARQL Matching |
	// +-----------------+

	public Future<Collection<IChunk>> findSPARQLMatches(final SPARQLQuery sparqlQuery) {
		return delayedFuture(new Callable<Collection<IChunk>>() {

			public Collection<IChunk> call() throws Exception {

				int retrievalID = getUniqueLabelID();
				RdfStoreConnection localStore = createRdfStoreConnection();

				URI resultLabel = new URIImpl("http://larkc.eu/jact-r/retrieval-result-" + retrievalID);
				localStore.associateStatements(sparqlQuery, resultLabel);

				return createSPARQLResultChunks(localStore, resultLabel);
			}

		}, getExecutor());
	}

	private int getUniqueLabelID() {
		int result = _uniqueLabelID;
		_uniqueLabelID++;
		return result;
	}

	protected List<IChunk> createSPARQLResultChunks(RdfStoreConnection localStore, URI inputLabel) {
		List<IChunk> result = new ArrayList<IChunk>();
		int counter = 0;

		CloseableIterator<Statement> chunks = localStore.search(null, RDF_TYPE_URI, null, null, inputLabel);
		while (chunks.hasNext()) {
			Statement statement = chunks.next();

			URI chunkLabel = new URIImpl(inputLabel.stringValue() + "-chunk-" + counter);
			counter++;

			localStore.associateStatements(statement.getSubject(), null, null, null, chunkLabel);
			LabelledGroupOfStatements chunkGroup = localStore.createLabelledGroupOfStatements(chunkLabel);

			String chunkName = addPrefix(statement.getSubject().stringValue());
			String chunkTypeName = addPrefix(statement.getObject().stringValue());

			IChunk newChunk = null;
			try {

				IChunkType chunkType = getChunkType(chunkTypeName).get();
				newChunk = createChunk(chunkType, chunkName).get();

			} catch (Exception e) {
				LOGGER.error("Unable to create chunk '" + chunkName + "'.", e);
			}

			if (newChunk != null) {
				((ISymbolicChunkRdf) newChunk.getSymbolicChunk()).setRdfData(chunkGroup);
				result.add(newChunk);
			}
		}
		chunks.close();

		return result;
	}

	// +----------------------+
	// | Parameter Management |
	// +----------------------+

	public RdfStoreConnection createRdfStoreConnection() {
		return DataFactory.INSTANCE.createRdfStoreConnection();
	}

	@Override	
	public boolean isPartialMatchingEnabled() {
		return false;
	}

	public String getRdfSourceURL() {
		return _sourceURL;
	}

	private void setRdfSourceURL( String sourceURL ) {
		_sourceURL = sourceURL;
	}

	public String getRdfFormat() {
		return _rdfFormat.toString();
	}

	private void setRdfFormat( String format ) {
		if ( format.equalsIgnoreCase( "RDF/XML" ) ) {
			_rdfFormat = RDFFormat.RDFXML;
		} else if ( format.equalsIgnoreCase( "N3" ) ) {
			_rdfFormat = RDFFormat.N3;
		} else if ( format.equalsIgnoreCase( "N-Triples" ) ) {
			_rdfFormat = RDFFormat.NTRIPLES;
		}
	}

	public String getBaseURI() {
		return _baseURI;
	}

	private void setBaseURI( String baseURI ) {
		_baseURI = baseURI;
	}

	public String getParameter(String key) {
		if (RDF_SOURCE.equalsIgnoreCase(key))
			return "" + getRdfSourceURL();
		if (BASE_URI.equalsIgnoreCase(key))
			return "" + getBaseURI();
		if (RDF_FORMAT.equalsIgnoreCase(key))
			return "" + getRdfFormat();
		return super.getParameter(key);
	}

	public Collection<String> getSetableParameters() {
		ArrayList<String> rtn = new ArrayList<String>();
		rtn.add(RDF_SOURCE);
		rtn.add(RDF_FORMAT);
		rtn.add(BASE_URI);
		return rtn;
	}

	public void setParameter(String key, String value) {
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Module Parameter set: "+ key +"="+ value);
		}

		if ( RDF_SOURCE.equalsIgnoreCase(key) )
			setRdfSourceURL( value );
		else if ( RDF_FORMAT.equalsIgnoreCase(key) )
			setRdfFormat( value );
		else if ( BASE_URI.equalsIgnoreCase(key) )
			setBaseURI( value );
		else 
			super.setParameter(key, value);
	}
}
