/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.modules.declarative.rdf;

import java.util.Collection;
import java.util.Map;
import java.util.concurrent.Future;

import org.jactr.core.chunk.IChunk;
import org.jactr.core.module.declarative.IDeclarativeModule;

import eu.larkc.core.data.RdfStoreConnection;
import eu.larkc.core.query.SPARQLQuery;

/**
 * Declarative modules that implement this interface can be used 
 * by LarCT-R modules and productions.
 * 
 * @author Martijn Brakenhoff
 * @created March 15, 2010
 */
public interface IDeclarativeModuleRdf extends IDeclarativeModule {
  
	public static final String RDF_SOURCE_FILE	= "RDFSourceFile";
	public static final String RDF_SOURCE_URL	= "RDFSourceURL";
	public static final String RDF_SOURCE		= "RDFSource";
	public static final String BASE_URI			= "BaseURI";
	public static final String RDF_FORMAT		= "RDFFormat";
	
	/**
	 * Returns the namespaces that are used in this module's 
	 * knowledgebase.
	 *  
	 * @return the namespaces.
	 */
	public Map<String, String> getNamespaces ();
	
	/**
	 * Returns the namespaces (see getNamespaces()), but formats 
	 * them like 
	 * the PREFIX part of a sparql query. 
	 * @return the prefix part of a sparql query.
	 */
	public String getSPARQLPrefixes ();
	
	/**
	 * Returns the resource, but with the namespace replaced by 
	 * the matching prefix.
	 * 
	 * @param resource the rdf resource URI.
	 * @return if the namespace is known, the resource with a 
	 * prefix, otherwise the given resource.
	 */
	public String addPrefix( String resource );
	
	/**
	 * Returns (the promise of) a collection of all the chunks 
	 * that match the given SPARQL Query.
	 * 
	 * @param sparqlQuery the SPARQL-Query object.
	 * @return all the chunks that match the given SPARQL Query.
	 */
	public Future<Collection<IChunk>> findSPARQLMatches(final SPARQLQuery sparqlQuery);
	
	/**
	 * Returns the RdfStoreConnection to the internal repository.
	 * 
	 * @return the RdfStoreConnection
	 */
	public RdfStoreConnection createRdfStoreConnection();
	
	/**
	 * Returns the URL that was set as the Source of RDF in the model.
	 * 
	 * @return the RDF source URL.
	 */
	public String getRdfSourceURL();

	/**
	 * Returns the RDF format that was set in the model. (Possible 
	 * values: String equivalent of org.openrdf.rio.RDFFormat). 
	 * 
	 * @return the RDF format.
	 */
	public String getRdfFormat();
	
	/**
	 * Returns the Base URI that was set for the knowledge base in 
	 * the model.
	 * 
	 * @return the Base URI for the knowledge base
	 */
	public String getBaseURI();
	
}