/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.modules.larkc.analyzer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jactr.core.chunk.IChunk;
import org.jactr.core.chunktype.IChunkType;
import org.jactr.core.module.retrieval.six.DefaultRetrievalModule6;
import org.jactr.core.production.request.ChunkTypeRequest;
import org.jactr.core.slot.ISlot;

import eu.larkc.core.data.InformationSet;
import eu.larkc.core.data.LabelledGroupOfStatements;
import eu.larkc.core.data.VariableBinding;
import eu.larkc.core.query.Query;
import eu.larkc.jactr.core.chunk.larkc.IChunkLarKC;
import eu.larkc.jactr.core.chunk.larkc.ISymbolicChunkLarKC;
import eu.larkc.jactr.core.chunk.rdf.DefaultChunkRdf;
import eu.larkc.jactr.core.chunk.rdf.IChunkRdf;
import eu.larkc.jactr.core.chunk.rdf.ISymbolicChunkRdf;

/**
 * The Module that allows easy analysis of LarKC results.
 * 
 * @author Martijn Brakenhoff
 * @created October 6, 2010
 */
public abstract class AbstractAnalyzerModule extends DefaultRetrievalModule6 {

	private static final transient Log 	LOGGER = LogFactory.getLog(AbstractAnalyzerModule.class);
	
	private IChunkType chunkTypeLarKC;
	private ArrayList<IChunk> outputQueue;
		
	public AbstractAnalyzerModule(String name) {
		super(name);
		resetOutputQueue();
	}
	
	@Override
	public Future<IChunk> retrieveChunk(final ChunkTypeRequest chunkPattern) {
		return delayedFuture(new Callable<IChunk>() {

			public IChunk call() throws Exception {
				startAnalysis(chunkPattern);
				
				if (outputQueue.size() > 0) {
					return outputQueue.get(0);
				} else {
					return getModel().getDeclarativeModule().getErrorChunk();
				}
			}
		}, getExecutor());
	}

	private void startAnalysis(ChunkTypeRequest chunkPattern) {
		Collection<ISlot> slots = chunkPattern.getSlots(new ArrayList<ISlot>());
		
		for (ISlot slot : slots) {
			if (slot.getName().equalsIgnoreCase(ISymbolicChunkLarKC.RDF_SLOT_NAME)) {
				analyze(slot.getValue());
			}
		}
	}
	
	private void analyze(Object value) {
		if (this instanceof IInformationSetAnalyzer && value instanceof InformationSet) {
			IInformationSetAnalyzer analyzer = ((IInformationSetAnalyzer)this);
			InformationSet infoSet = (InformationSet) value;
			
			analyzer.analyze(infoSet);
		} else if (this instanceof IVariableBindingAnalyzer && value instanceof VariableBinding) {
			IVariableBindingAnalyzer analyzer = ((IVariableBindingAnalyzer)this);
			VariableBinding infoSet = (VariableBinding) value;
			
			analyzer.analyze(infoSet);
		}
	}

	protected IChunkRdf createRdfChunk (IChunkType chunkType, LabelledGroupOfStatements chunkData) {
		DefaultChunkRdf result = new DefaultChunkRdf(chunkType);
		ISymbolicChunkRdf symbolicChunk = (ISymbolicChunkRdf)result.getSymbolicChunk();
		
		symbolicChunk.setRdfData(chunkData);
		
		return result;
	}

	protected IChunkRdf createRdfChunk (String chunkTypeName, LabelledGroupOfStatements chunkData) {
		IChunkType ct = null;
		try {
			ct = getModel().getDeclarativeModule().getChunkType(chunkTypeName).get();

		} catch (InterruptedException e) {
			e.printStackTrace();
			LOGGER.error("Exception while finding chunktype: "+ chunkTypeName, e);
		} catch (ExecutionException e) {
			e.printStackTrace();
			LOGGER.error("Exception while finding chunktype: "+ chunkTypeName, e);
		}
		
		if (ct == null) {
			throw new IllegalArgumentException("Unknown ChunkType: "+ chunkTypeName); 
		}
		return createRdfChunk(ct, chunkData);
	}
	
	protected IChunkLarKC createLarKCChunk (InformationSet larkcData) {
		DefaultChunkRdf result = new DefaultChunkRdf(chunkTypeLarKC);
		ISymbolicChunkLarKC symbolicChunk;
		
		result.enableLarKCData();
		symbolicChunk = (ISymbolicChunkLarKC)result.getSymbolicChunk();
		symbolicChunk.setLarKCData(larkcData);
		
		return result;
	}

	protected IChunkLarKC createLarKCChunk (Query larkcData) {
		DefaultChunkRdf result = new DefaultChunkRdf(chunkTypeLarKC);
		ISymbolicChunkLarKC symbolicChunk;
		
		result.enableLarKCData();
		symbolicChunk = (ISymbolicChunkLarKC)result.getSymbolicChunk();
		symbolicChunk.setLarKCData(larkcData);
		
		return result;	
	}

	protected IChunkLarKC createLarKCChunk (Collection<?> larkcData) {
		DefaultChunkRdf result = new DefaultChunkRdf(chunkTypeLarKC);
		ISymbolicChunkLarKC symbolicChunk;
		
		result.enableLarKCData();
		symbolicChunk = (ISymbolicChunkLarKC)result.getSymbolicChunk();
		symbolicChunk.setLarKCDataCollection(larkcData);
		
		return result;
	}

	protected void putOnOutputQueue (IChunk chunk) {
		outputQueue.add(chunk);
	}

	protected void resetOutputQueue () {
		outputQueue = new ArrayList<IChunk>();
	}

}