/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.scripting.sparql;

import java.util.Collections;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jactr.core.buffer.IActivationBuffer;
import org.jactr.core.buffer.IRequestableBuffer;
import org.jactr.core.chunktype.IChunkType;
import org.jactr.core.model.IModel;
import org.jactr.core.module.IModule;
import org.jactr.core.module.declarative.IDeclarativeModule;
import org.jactr.core.production.CannotInstantiateException;
import org.jactr.core.production.IInstantiation;
import org.jactr.core.production.VariableBindings;
import org.jactr.core.production.request.ChunkTypeRequest;
import org.jactr.core.slot.DefaultMutableSlot;
import org.jactr.core.utils.ModelerException;
import org.jactr.scripting.IScriptableFactory;
import org.jactr.scripting.ScriptSupport;
import org.jactr.scripting.action.IActionScript;

import eu.larkc.core.query.SPARQLQuery;
import eu.larkc.core.query.SPARQLQueryImpl;
import eu.larkc.jactr.modules.declarative.rdf.IDeclarativeModuleRdf;
import eu.larkc.jactr.modules.larkc.plugin.ILarKCPluginModule;
import eu.larkc.jactr.modules.sparql.buffer.DefaultSPARQLRetrievalBuffer;

/**
 * Scriptable Action that represents a SPARQL request.
 * 
 * @author Martijn Brakenhoff
 * @created March 25, 2010
 */
public class SPARQLAction implements IActionScript {
	/**
	 * Logger definition
	 */
	private static final transient Log 		LOGGER = LogFactory.getLog(SPARQLAction.class);

	private static final String 			LARKC_DATA_CHUNKTYPE_NAME = "LarKCData";

	private final IScriptableFactory  		 _factory;

	private final String               		_script;

	private final SPARQLQueryInstantiation	_completedQuery;

	private SPARQLQuery 					_compiledQuery;

	public SPARQLAction(String script, IScriptableFactory factory) {
		_factory = factory;
		_script = script;

		_completedQuery = new SPARQLQueryInstantiation(script, "CONSTRUCT");
	}

	public IActionScript bind(VariableBindings variableBindings)
	throws CannotInstantiateException {
		Map<String, String> namespaces = null;

		Object model = variableBindings.get("=model");
		if (model instanceof IModel) {
			IDeclarativeModule dm = ((IModel) model).getDeclarativeModule();
			if (dm instanceof IDeclarativeModuleRdf) {
				namespaces = ((IDeclarativeModuleRdf) dm).getNamespaces();
				_completedQuery.addPrefixes(namespaces);
			} else {
				throw new ModelerException("Error in Scriptable IAction", 
				"Unable to perform SPARQL actions if the Declarative Module does not support RDF.");
			}
		}

		try {
			_compiledQuery = new SPARQLQueryImpl(_completedQuery.getBoundQuery(variableBindings));
		} catch (Exception ioe) {
			LOGGER.error("Error in scriptable action", ioe);
			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Query: " + _completedQuery.getBoundQuery(variableBindings));
			}

			throw new ModelerException("Error in Scriptable IAction: ", ioe,
			"double check your SPARQL sytanx. The SPARQL parser detected an error.");
		}

		return this;
	}

	public double fire(ScriptSupport scriptSupport, IInstantiation instantiation, 
			double firedAt) {
		IModel model = scriptSupport.getModel();
		IActivationBuffer buffer = model.getActivationBuffer(_completedQuery.getBufferName());
		IModule module = buffer.getModule();
		
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("SPARQL Query Fired to module: " + module.getName());
		}

		if (buffer instanceof DefaultSPARQLRetrievalBuffer) {
			SPARQLRequest request = new SPARQLRequest(_compiledQuery);
			((DefaultSPARQLRetrievalBuffer) buffer).request(request, firedAt);

		} else if (buffer instanceof IRequestableBuffer && module instanceof ILarKCPluginModule){

			IChunkType larkcDataChunkType = null;
			try {

				larkcDataChunkType = instantiation.getModel().getDeclarativeModule().getChunkType(LARKC_DATA_CHUNKTYPE_NAME).get();

			} catch (Exception e) {
				LOGGER.error("Thread Error while firing SPARQLAction: Can't get the LarKCData ChunkType.", e);				
			}

			if (larkcDataChunkType != null) {
				if (LOGGER.isDebugEnabled()) {
			    	LOGGER.debug("SPARQL Query Fired to module: " + module.getName());
				}

				ChunkTypeRequest request = new ChunkTypeRequest(larkcDataChunkType, Collections.singleton(new DefaultMutableSlot("rdf", _compiledQuery)));
				((IRequestableBuffer) buffer).request(request, firedAt);					
			}
		}

		return 0.05;
	}

	public IScriptableFactory getFactory() {
		return _factory;
	}

	public String getScript() {
		return _script;
	}

	public void dispose() {
		// noop since we recycle
	}
}
