/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.scripting.sparql;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jactr.core.chunk.IChunk;
import org.jactr.core.production.VariableBindings;

/**
 * 
 * 
 * @author Martijn Brakenhoff
 * @created April 1, 2010
 */
public class SPARQLQueryInstantiation {
	
	private static final transient Log 	LOGGER = LogFactory.getLog(SPARQLQueryInstantiation.class);
	
	private static final String URL_REGEX = "[\\w\\d:#@%/;\\$()~_\\?\\+-=\\\\.&]*";

	private static final String ANY_CHAR_PATTERN = "[\\w\\s:\\?\\$\\*\"\\^\\n/#@%;~_\\+-=&\\.<>\\\\]";

	private static final Pattern BUFFER_PATTERN  =   Pattern.compile("BUFFER[\\s]*\\([\\w\\d\\s]+\\)");

	private static final Pattern BUFFER_PATTERN_2  = Pattern.compile("BUFFER[\\s]*\\([\\w\\d\\s]+\\)[\\n]*");
	
	private static final Pattern BUFFER_PATTERN_3  = Pattern.compile("[\\?\\$][\\w\\d]+");
	
	private static final Pattern BUFFERNAME  = Pattern.compile("[^(BUFFER)][^(\\s)][^(\\()][^(\\s)][\\w\\s]+[^(\\s)][^(\\))]");
	
	private static final Pattern PREFIX_LINE_PATTERN = Pattern.compile("PREFIX[\\s]+[\\w]+:[\\s]+<http://"+ URL_REGEX +">");
	


	private static final String NO_BUFFER_SPECIFIED = "A SPARQL script must specify a Buffer on the first line: BUFFER ( buffername )";

	private static final String NON_AUTOCOMPLETE_TYPE = "A SPARQLQuery can ony be auto-completed for SELECT or CONSTRUCT queries.";

	private static final String SELECT = "SELECT";

	private static final String CONSTRUCT = "CONSTRUCT";

	private static final String MUST_CONTAIN_PATTERN_OR_QUERY = "A SPARQL script must contain either a Triple Pattern or a full query.";

	private static final String FIND_FILTERS = "FILTER[\\s]*\\(" + ANY_CHAR_PATTERN + "*\\)[\\s\\.]*";
	
	private String  _type,
					_bufferName,
					_complementedQuery;
	private Map<String, String> _prefixes;

	public SPARQLQueryInstantiation(String script, String autoCompleteType) {
		_type = autoCompleteType;
		
	    _bufferName  = findBufferName( script );
	    _prefixes    = findPrefixes( script );
	    
	    _complementedQuery = findQuery( script );
	}
	
	protected String findBufferName( String script ) {
		Matcher matcher = BUFFER_PATTERN.matcher(script);
		
		String bufferLine = null;
        while(matcher.find()) {
               bufferLine = matcher.group().trim();
        }
        
        if (bufferLine != null) {
        	matcher = BUFFERNAME.matcher(bufferLine);
        	while(matcher.find()) {
        		return matcher.group().trim();
        	}
        }
        throw new IllegalArgumentException(NO_BUFFER_SPECIFIED);
	}

	protected Map<String, String> findPrefixes(String script) {
		Map<String, String> result = new HashMap<String, String>();
		
		Matcher matcher = PREFIX_LINE_PATTERN.matcher(script);
		
		while(matcher.find()) {
       		Scanner scanName = new Scanner(matcher.group());
       		String prefix = scanName.findInLine("[^(PREFIX)][^(\\s)][\\w]+").trim();
        		
       		Scanner scanURL = new Scanner(matcher.group());
       		String prefixURL = scanURL.findInLine("http://"+ URL_REGEX);
        		
       		result.put(prefix, prefixURL);
        }
        
        return result;
	}
	
	protected String findQuery(String script) {
		String result = findSelectQuery( script );
	    // if it's not already a full select query, then try to find a construct
	    if ( result == null) {
	    	 result = findConstructQuery( script );
	    }
	    
	    // if it's not already a full query, then use the autoCompleteType
	    if ( result == null) {
	    	String triplePattern = findTriplePattern(script);
	    	if (isSelect()) {
	    		 result = addSelectStatement( triplePattern );
	    	} else if (isConstruct()) {
	    		 result = addConstructStatement( triplePattern );
	    	} else {
	    		throw new IllegalArgumentException(NON_AUTOCOMPLETE_TYPE); 
	    	}
	    }
		return result;
	}
	
	protected String findSelectQuery(String script) {
		final Pattern selectQueryPattern  =
			Pattern.compile(SELECT);	
		Matcher matcher = selectQueryPattern.matcher(script);
		
		StringBuffer query = new StringBuffer();
		
		if (matcher.find()) {
			int end = matcher.end();
			query.append( matcher.group() );
			
			query.append( script.substring(end) );
			
			return query.toString();
		} else {
			return null;
		}
	}
	
	protected String findConstructQuery(String script) {
		final Pattern constructQueryPattern  =
			Pattern.compile(CONSTRUCT);	
		Matcher matcher = constructQueryPattern.matcher(script);
		
		StringBuffer query = new StringBuffer();
		
		if (matcher.find()) {
			int end = matcher.end();
			query.append( matcher.group() );
			
			query.append( script.substring(end) );
			
			return query.toString();
		} else {
			return null;
		}
	}
	
	protected String findTriplePattern( String script ) {
		Matcher matcher = BUFFER_PATTERN_2.matcher(script);
		
        while(matcher.find()) {
        	int index = matcher.end() + 1;
        	if (script.length() > index) {
        		return script.substring( index ).trim();
        	}
        }
        throw new IllegalArgumentException(MUST_CONTAIN_PATTERN_OR_QUERY);
	}
	
	protected String addSelectStatement(String triplePattern) {
		StringBuffer result = new StringBuffer();

		Matcher matcher = BUFFER_PATTERN_3.matcher(triplePattern);
		
		Set<String> variables = new HashSet<String>();
        while(matcher.find()) {
        	variables.add(matcher.group(0));
        }
        
        result.append("SELECT ");
        for (String var : variables) {
        	result.append(var + " ");
        }
        result.append(" WHERE {\r\n");
        result.append(triplePattern);
        result.append("\r\n}");
        
        return result.toString();
	}
	
	protected String addConstructStatement(String triplePattern) {
		StringBuffer result = new StringBuffer();

        result.append("CONSTRUCT {\r\n");
		// remove Filters from first triple pattern.
        result.append(cleanTriplePattern(triplePattern));
        result.append("\r\n} WHERE {\r\n");
        result.append(triplePattern);
        result.append("\r\n}");
        
        return result.toString();
	}

	private Object cleanTriplePattern(String triplePattern) {
		return triplePattern.replaceAll(FIND_FILTERS, "");
	}

	/**
	 * Checks if the auto-complete type is "Select".
	 * 
	 * @return true if the auto-complete type is "Select", otherwise false.
	 */
	public boolean isSelect() {
		return _type.equalsIgnoreCase(SELECT);
	}
	
	/**
	 * Checks if the auto-complete type is "Construct".
	 * 
	 * @return true if the auto-complete type is "Construct", otherwise false.
	 */
	public boolean isConstruct() {
		return _type.equalsIgnoreCase(CONSTRUCT);
	}
	
	/**
	 * Returns all the prefixes that were set.
	 * 
	 * @return the prefixes
	 */
	public Map<String, String> getPrefixes() {
		return _prefixes;
	}

	/**
	 * Adds the given prefixes.
	 * 
	 * @param prefixes the prefixes that should be added.
	 */
	public void addPrefixes(Map<String, String> prefixes) {
		for (String key : prefixes.keySet()) {
			if (!_prefixes.containsKey(key)) {
				_prefixes.put(key, prefixes.get(key));
			}
		}
	}

	/**
	 * Returns the buffername, that was found in the script. 
	 * 
	 * @return the buffername.
	 */
	public String getBufferName() {
		return _bufferName;
	}
	
	private String getPrefixString() {
		StringBuffer result = new StringBuffer();
		for (String prefix : getPrefixes().keySet()) {
			result.append("PREFIX " + prefix + ": <" + getPrefixes().get(prefix)+">\r\n");
		}
		return result.toString();
	}

	/**
	 * Returns the auto-completed query-string.
	 * 
	 * @return the query-string
	 */
	public String getQuery() {
		return getPrefixString() + _complementedQuery;
	}
	
	/**
	 * Returns the query-string, but replaces the variables 
	 * with the objects from the bindings.
	 * 
	 * @param variableBindings the variable values.
	 * @return the bounded query.
	 */
	public String getBoundQuery (VariableBindings variableBindings) {
		Set<String> boundVariables = variableBindings.getVariables();
		String boundQuery = _complementedQuery;
		
		for (String key : boundVariables) {
			Object value = variableBindings.get( key );
			if (value instanceof IChunk) {
				value = ((IChunk) value).getSymbolicChunk().getName();
			}
			if (value instanceof Integer) {
				value = ((Integer) value).toString() + "^^xsd:integer";
			}
			if (value instanceof Double) {
				value = "\"" + ((Double) value).toString() + "\"^^xsd:double";
			}
			if (value instanceof String) {
				if (LOGGER.isTraceEnabled()) {
					LOGGER.trace("Checking bound variables: "+ key +"="+ value);
				}
				boundQuery = boundQuery.replaceAll("\\?"+ key.substring(1), (String) value)
									   .replaceAll("\\$"+ key.substring(1), (String) value);
			}
		}
		
		if (isSelect()) {
			boundQuery = cleanVariableList(boundQuery);
		}
		
		return getPrefixString() + boundQuery;
	}

	private String cleanVariableList(String query) {
		// Finds the SELECT-WHERE line(s) and replaces them with a clean version.
		final Pattern selectPattern = Pattern.compile("SELECT "+ANY_CHAR_PATTERN + "+ WHERE");
		Matcher matcher = selectPattern.matcher(query);
		
		if (matcher.find()) {
			String replacement = cleanVariableListFromSelectLine(matcher.group());
			return query.substring(0, matcher.start()) + replacement + query.substring(matcher.end());
		}
		return query;
	}

	private String cleanVariableListFromSelectLine(String line) {
		StringBuffer result = new StringBuffer();
		Scanner scanner = new Scanner(line).useDelimiter("\\s");
		
		if (scanner.hasNext()) {
			String match = scanner.next().trim();
			
			while(match.equals("SELECT") || match.equals("DISTINCT") || match.equals("REDUCED") || match.isEmpty()) {
				result.append(match + " ");
				
				match = scanner.next().trim();
			}
			
			while(!match.equals("FROM") && !match.equals("WHERE")) {
				if (match.isEmpty()) {
					result.append(" ");
				} else if (match.startsWith("?") ||match.startsWith("$") || match.startsWith("*")) {
					result.append(match + " ");
				}

				match = scanner.next().trim();
			}

			result.append(match + " ");
			while(scanner.hasNext()) {
				match = scanner.next();
				result.append(match + " ");
			}
		} 
		
		return result.toString();
	}
}
